# "A Bootstrap Method for Conducting Statistical Inference with Clustered Data"
#
# Jeffrey J. Harden
# University of North Carolina at Chapel Hill
# jjharden@unc.edu
#
# last update: May 23, 2010
#
################################################################################
# Table of contents:
#  PART I: Loading packages
#  PART II: Creating the necessary functions
#  PART III: Monte Carlo simulations
#
# Requires packages mvtnorm, Design, Hmisc, and sandwich
#
# This code is based on the Stata code used in:
#  Green, Donald P. and Lynn Vavreck. 2008. "Analysis of Cluster-Randomized Experiments:
#   A Comparison of Alternative Estimation Techniques." Political Analysis 16(2):138-152.
#
################################################################################
# PART I: Loading packages
library(mvtnorm)
library(Design)
library(Hmisc)
library(sandwich)
#
################################################################################
# PART II: Creating the necessary functions
# RCSE function (based on code from Mahmood Arai, http://people.su.se/~ma/clustering.pdf)
cluster.se <- function(model, cluster){
         M <- length(unique(cluster))
         N <- length(cluster)
         K <- model$rank
         dfc <- (M/(M - 1)) * ((N - 1)/(N - K))
         uj  <- apply(estfun(model), 2, function(x) tapply(x, cluster, sum));
         rcse.cov <- dfc * sandwich(model, meat = crossprod(uj)/N)
         return(rcse.cov)
        }

# RCSE function that uses the true errors instead of the residuals
cluster.se.true <- function(model, epsilon, cluster){
         M <- length(unique(cluster))
         N <- length(cluster)
         K <- model$rank
         dfc <- (M/(M - 1)) * ((N - 1)/(N - K))
         uj  <- apply((epsilon * model$x), 2, function(x) tapply(x, cluster, sum));
         rcse.cov <- dfc * sandwich(model, meat = crossprod(uj)/N)
         return(rcse.cov)
        }

# 95% CI coverage probability function (code from Bruce Desmarais)
comp <- function(bse, p){
return(as.numeric(p >= bse[1] - 1.96 * bse[2] & p <= bse[1] + 1.96 * bse[2]))
}
coverage <- function(bse, par){
return(mean(apply(bse, 1, comp, p = par)))
}

################################################################################
# PART III: Monte Carlo simulations
#  Control panel
set.seed(82184)                     
n <- 800 # Sample size
p <- 0.5 # Rho
nc <- 25 # Number of clusters
simulations <- 10000 # Number of Monte Carlo simulations
boot <- 1000 # Number of bootstrap replications

a <- 0 # True parameter values
beta1 <- 0.85
beta2 <- 0.5
beta3 <- 0.7
cluster <- rep(1:nc, each = n/nc) # Cluster label

B1 <- numeric(simulations) # Vectors to store estimates
B2 <- numeric(simulations)
B3 <- numeric(simulations)
SE1 <- numeric(simulations)
SE2 <- numeric(simulations)
SE3 <- numeric(simulations)
RCSE1 <- numeric(simulations)
RCSE2 <- numeric(simulations)
RCSE3 <- numeric(simulations)
RCSE.t1 <- numeric(simulations)
RCSE.t2 <- numeric(simulations)
RCSE.t3 <- numeric(simulations)
BCSE1 <- numeric(simulations)
BCSE2 <- numeric(simulations)
BCSE3 <- numeric(simulations)
rho.hat <- numeric(simulations)
ratio <- numeric(simulations)
mean.e <- numeric(simulations)
mean.r <- numeric(simulations)

  for (i in 1:simulations){ # Simulate a clustered data set, estimate OLS model

  c.sigma <- matrix(c(4, 0, 0, p), ncol = 2) # Cluster-level random effects
  c.values <- rmvnorm(n = nc, sigma = c.sigma)
  randeff1 <- rep(c.values[ , 1], each = n/nc)
  randeff2 <- rep(c.values[ , 2], each = n/nc)

  i.sigma <- matrix(c(1, 0, 0, (1 - p)), ncol = 2) # Individual-level random effects
  i.values <- rmvnorm(n = n, sigma = i.sigma)
  randeff3 <- i.values[ , 1]
  randeff4 <- i.values[ , 2]

  X1 <- 3 + randeff1 + randeff3 # X1 values unique to individual observations
  X2 <- randeff1 # X2 values unique to clusters of observations
  epsilon <- randeff2 + randeff4 # Two components of the error term

  Y <- a + beta1 * X1 + beta2 * X2 + beta3 * (X1 * X2) + epsilon # True model

  fit.ols <- ols(Y ~ X1 + X2 + X1*X2, x = TRUE, y = TRUE) # Model fitting
  rcse <- cluster.se(fit.ols, cluster) # RCSE covariance matrix
  rcse.t <- cluster.se.true(fit.ols, epsilon, cluster) # RCSE covariance matrix with true errors
  bcse <- bootcov(fit.ols, cluster, B = boot) # BCSE covariance matrix

  B1[i] <- fit.ols$coef[2] # Collect coefficients and standard errors
  B2[i] <- fit.ols$coef[3]
  B3[i] <- fit.ols$coef[4]
  SE1[i] <- sqrt(fit.ols$var[2, 2])
  SE2[i] <- sqrt(fit.ols$var[3, 3])
  SE3[i] <- sqrt(fit.ols$var[4, 4])
  RCSE1[i] <- sqrt(rcse[2, 2])
  RCSE2[i] <- sqrt(rcse[3, 3])
  RCSE3[i] <- sqrt(rcse[4, 4])
  RCSE.t1[i] <- sqrt(rcse.t[2, 2])
  RCSE.t2[i] <- sqrt(rcse.t[3, 3])
  RCSE.t3[i] <- sqrt(rcse.t[4, 4])
  BCSE1[i] <- sqrt(bcse$var[2, 2])
  BCSE2[i] <- sqrt(bcse$var[3, 3])
  BCSE3[i] <- sqrt(bcse$var[4, 4])

  rho.hat[i] <- deff(fit.ols$residuals, cluster)[3] # Calculate estimate of rho
    if (rho.hat[i] < 0){
      rho.hat[i] <- 0
    }
  ratio[i] <- var(fit.ols$residuals)/var(epsilon)
  mean.e[i] <- mean(epsilon)
  mean.r[i] <- mean(fit.ols$residuals)
}

cover1 <- coverage(cbind(B1, SE1), beta1) # Coverage proportion
cover2 <- coverage(cbind(B2, SE2), beta2)
cover3 <- coverage(cbind(B3, SE3), beta3)
rccover1 <- coverage(cbind(B1, RCSE1), beta1)
rccover2 <- coverage(cbind(B2, RCSE2), beta2)
rccover3 <- coverage(cbind(B3, RCSE3), beta3)
rccover.t1 <- coverage(cbind(B1, RCSE.t1), beta1)
rccover.t2 <- coverage(cbind(B2, RCSE.t2), beta2)
rccover.t3 <- coverage(cbind(B3, RCSE.t3), beta3)
bccover1 <- coverage(cbind(B1, BCSE1), beta1)
bccover2 <- coverage(cbind(B2, BCSE2), beta2)
bccover3 <- coverage(cbind(B3, BCSE3), beta3)

mean1 <- mean(B1) # Coefficient means
mean2 <- mean(B2)
mean3 <- mean(B3)
sd1 <- sd(B1) # Coefficient standard deviations
sd2 <- sd(B2)
sd3 <- sd(B3)

meanse1 <- mean(SE1) # Standard error means
meanse2 <- mean(SE2)
meanse3 <- mean(SE3)
meanrcse1 <- mean(RCSE1)
meanrcse2 <- mean(RCSE2)
meanrcse3 <- mean(RCSE3)
meanrcse.t1 <- mean(RCSE.t1)
meanrcse.t2 <- mean(RCSE.t2)
meanrcse.t3 <- mean(RCSE.t3)
meanbcse1 <- mean(BCSE1)
meanbcse2 <- mean(BCSE2)
meanbcse3 <- mean(BCSE3)

avg.rho.hat <- mean(rho.hat)
avg.ratio <- mean(ratio)
avg.mean.e <- mean(mean.e)
avg.mean.r <- mean(mean.r)

results <- cbind(cover1, rccover1, rccover.t1, bccover1, cover2, rccover2, rccover.t2, bccover2, cover3, rccover3, rccover.t3, bccover3, mean1, mean2, mean3, sd1, meanse1, meanrcse1, meanrcse.t1, meanbcse1, sd2, meanse2, meanrcse2, meanrcse.t2, meanbcse2, sd3, meanse3, meanrcse3, meanrcse.t3, meanbcse3, avg.rho.hat, avg.ratio, avg.mean.e, avg.mean.r)

write.csv(results, "n800p5c25.csv", row.names = nc)